// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package repo

import (
	"context"
	"database/sql"
	"errors"
	"time"

	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/adapter/repo/sqlc"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/model"
	"git.vinchent.xyz/vinchent/howmuch/internal/howmuch/usecase/repo"
	"github.com/jackc/pgx/v5"
)

type userRepository struct {
	db *sql.DB
}

const insertTimeout = 1 * time.Second

func NewUserRepository(db *sql.DB) repo.UserRepository {
	return &userRepository{
		db: db,
	}
}

// Create
func (ur *userRepository) Create(
	ctx context.Context,
	transaction interface{},
	u *model.User,
) (*model.User, error) {
	timeoutCtx, cancel := context.WithTimeout(ctx, insertTimeout)
	defer cancel()

	args := sqlc.InsertUserParams{
		Email:     u.Email,
		FirstName: u.FirstName,
		LastName:  u.LastName,
		Password:  u.Password,
		CreatedAt: time.Now(),
		UpdatedAt: time.Now(),
	}

	tx, ok := transaction.(*sql.Tx)
	if !ok {
		return nil, errors.New("transaction is not a pgx.Tx")
	}

	queries := sqlc.New(tx)

	userDB, err := queries.InsertUser(timeoutCtx, args)
	if err != nil {
		return nil, err
	}

	return &model.User{
		ID:        int(userDB.ID),
		Email:     userDB.Email,
		FirstName: userDB.FirstName,
		LastName:  userDB.LastName,
		Password:  userDB.Password,
		CreatedAt: userDB.CreatedAt,
		UpdatedAt: userDB.CreatedAt,
	}, nil
}

// GetByEmail if not found, return nil for user but not error.
func (ur *userRepository) GetByEmail(ctx context.Context, email string) (*model.User, error) {
	queries := sqlc.New(ur.db)
	userDB, err := queries.GetUserByEmail(ctx, email)
	if errors.Is(err, pgx.ErrNoRows) {
		// No query error, but user not found
		return nil, nil
	} else if err != nil {
		return nil, err
	}

	return &model.User{
		ID:        int(userDB.ID),
		Email:     userDB.Email,
		FirstName: userDB.FirstName,
		LastName:  userDB.LastName,
		Password:  userDB.Password,
		CreatedAt: userDB.CreatedAt,
		UpdatedAt: userDB.CreatedAt,
	}, nil
}
