// MIT License
//
// Copyright (c) 2024 vinchent <vinchent@vinchent.xyz>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

package datastore

import (
	"database/sql"
	"time"

	"git.vinchent.xyz/vinchent/howmuch/internal/pkg/log"
	_ "github.com/jackc/pgx/v5"
	_ "github.com/jackc/pgx/v5/stdlib"
)

var counts int

type DbExtraConf struct {
	MaxOpenConns int
	MaxIdleConns int
	MaxLifetime  time.Duration
}

// NewDB creates a new database for the application
func NewDB(dsn string, opts interface{}) *sql.DB {
	var db *sql.DB
	var err error

	for {
		db, err = openDB(dsn)
		if err != nil {
			log.WarnLog("postgres not ready", "err", err)
			counts++
		} else {
			log.InfoLog("connected to postgres")
			break
		}

		if counts > 10 {
			break
		}
		log.InfoLog("retry in 2 seconds")
		time.Sleep(2 * time.Second)
	}

	if db == nil {
		return nil
	}

	extraConf, ok := opts.(DbExtraConf)
	if ok {
		db.SetMaxOpenConns(extraConf.MaxOpenConns)
		db.SetMaxIdleConns(extraConf.MaxIdleConns)
		db.SetConnMaxLifetime(extraConf.MaxLifetime)
	}
	return db
}

func openDB(dsn string) (*sql.DB, error) {
	db, err := sql.Open("pgx", dsn)
	if err != nil {
		return nil, err
	}

	if err := db.Ping(); err != nil {
		return nil, err
	}
	return db, nil
}
